/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.fieldassist;

import org.eclipse.jface.fieldassist.ContentProposalAdapter;
import org.eclipse.jface.fieldassist.IContentProposal;
import org.eclipse.jface.fieldassist.IContentProposalListener;
import org.eclipse.jface.fieldassist.SimpleContentProposalProvider;
import org.eclipse.jface.fieldassist.TextContentAdapter;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Text;

/**
 * This class provides easy access to auto completion in {@link Text} fields.
 * Unlike the original this implementation takes care of proper in place
 * insertion of proposals and provides ability to dynamically provide context
 * sensitive proposals.
 * 
 * @author doebber
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 13B6BCCFF3C40E4D8D23A304DF26CE4A
 */
public class FieldAssist {

	/** The text field to assist. */
	private final Text field;

	/**
	 * Stores the custom proposal provider implemented for the concrete
	 * application.
	 */
	private IProposalProvider concreteProvider;

	/**
	 * ProposalProvider wrapper class that takes care of proper insertion. This
	 * is intended to be transparent to the implementor.
	 */
	private AF3ContentProposalProvider provider;

	/** Stores the {@link ContentProposalAdapter}. */
	private ContentProposalAdapter adapter;

	/**
	 * Constructor.
	 */
	public FieldAssist(Text field, IProposalProvider provider) {
		this.field = field;
		this.provider = new AF3ContentProposalProvider(provider);
		initialize();
	}

	/** Initializes the FieldAssist object. */
	private void initialize() {
		provider.setFiltering(true);
		adapter =
				new ContentProposalAdapter(field, new TextContentAdapter(), provider,
						concreteProvider.getKeystroke(), concreteProvider.getAutoActivationChars());
		adapter.setPropagateKeys(true);
		adapter.setProposalAcceptanceStyle(ContentProposalAdapter.PROPOSAL_IGNORE);

		adapter.setLabelProvider(new LabelProvider() {

			/** {@inheritDoc} */
			@Override
			public String getText(Object element) {
				if(element instanceof IContentProposal) {
					IContentProposal prop = (IContentProposal)element;
					if(prop.getLabel() == null) {
						return prop.getContent();
					}
					return prop.getLabel();
				}
				return element.toString();
			}

			/** {@inheritDoc} */
			@Override
			public Image getImage(Object element) {
				return FieldAssistImageRegistry.getIcon(getText(element));
			}
		});
		adapter.addContentProposalListener(new IContentProposalListener() {

			@Override
			/** {@inheritDoc} */
			public void proposalAccepted(IContentProposal proposal) {
				int end = field.getCaretPosition();
				int start = findBeginningOfWord(field.getText(), end);
				String newContents = field.getText().substring(0, Math.max(0, start + 1));
				newContents = newContents.concat(proposal.getContent());
				int min = Math.min(field.getText().length(), end);
				newContents = newContents.concat(field.getText().substring(min));
				field.setText(newContents);
				field.setSelection(start + proposal.getContent().length() + 1);
			}
		});
	}

	/**
	 * Sets a new proposal provider. Useful if the {@link FieldAssist} object is
	 * not disposed as its enclosing text field is not disposed.
	 */
	public void setProposalProvider(IProposalProvider proposalProvider) {
		this.provider = new AF3ContentProposalProvider(proposalProvider);
		provider.setFiltering(true);
		adapter.setContentProposalProvider(provider);
	}

	/**
	 * Wrapper class transparent to the implementor that takes care of proper in
	 * place insertion.
	 */
	private final class AF3ContentProposalProvider extends SimpleContentProposalProvider {

		/**
		 * Constructor for static proposals. Is never called but required by the
		 * compiler.
		 */
		public AF3ContentProposalProvider(String[] proposals) {
			super(proposals);
			concreteProvider = new StaticProposalProvider(proposals);
		}

		/** Constructor. */
		public AF3ContentProposalProvider(IProposalProvider provider) {
			super(new String[0]);
			concreteProvider = provider;
		}

		/** {@inheritDoc} */
		@Override
		public IContentProposal[] getProposals(String contents, int position) {
			int tmp = findBeginningOfWord(contents, position);
			String currentWord = contents.substring(tmp + 1, position);
			setProposals(concreteProvider.getProposals(currentWord));
			return super.getProposals(currentWord, position);
		}
	}

	/** Finds the beginning of the word the cursor is placed in. */
	private int findBeginningOfWord(String contents, int position) {
		if(contents.isEmpty()) {
			return -1;
		}
		int tmp = position - 1;
		while(tmp >= 0 && !Character.isWhitespace(contents.charAt(tmp)) &&
				!isAdditionalWhitespace(contents.charAt(tmp))) {
			tmp--;
		}
		return tmp;
	}

	/** Tests for match with additional whitespace characters. */
	private boolean isAdditionalWhitespace(char charAt) {
		for(char c : concreteProvider.getAdditionalWhitespaceChars()) {
			if(c == charAt) {
				return true;
			}
		}
		return false;
	}
}
