/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2013 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.library;

import static org.conqat.ide.commons.ui.dialog.MessageUtils.showErrorAsync;
import static org.fortiss.tooling.base.ui.library.LibrarySelectionDialog.openLibraryPackageDialog;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.ActionContributionItem;
import org.eclipse.jface.action.IContributionItem;
import org.fortiss.tooling.base.library.ModelElementLibraryService;
import org.fortiss.tooling.base.ui.ToolingBaseUIActivator;
import org.fortiss.tooling.kernel.model.ILibraryElementReference;
import org.fortiss.tooling.kernel.model.ILibraryPackage;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.ui.extension.IContextMenuContributor;
import org.fortiss.tooling.kernel.ui.extension.data.ContextMenuContextProvider;
import org.fortiss.tooling.kernel.ui.service.IContextMenuService;

/**
 * Base class for adding elements to a library.
 * 
 * @param <T>
 *            the type of the element to be added to the library.
 * 
 * @author ratiu
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 6B08DD35BE3A1B333418F5FA0004AD98
 */
public abstract class AddElementToLibraryBase<T extends EObject> implements IContextMenuContributor {

	/** {@inheritDoc} */
	@Override
	public List<IContributionItem> getContributedItems(EObject obj,
			ContextMenuContextProvider contextProvider) {
		boolean isObjLibRef = obj instanceof ILibraryElementReference;
		if(obj != null && getElementClass().isAssignableFrom(obj.getClass()) && !isObjLibRef) {
			AddToLibraryAction action = new AddToLibraryAction(obj);
			List<IContributionItem> contributionItems = new ArrayList<IContributionItem>();
			contributionItems.add(new ActionContributionItem(action));
			return contributionItems;
		}
		return Collections.emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public String getMenuSectionID() {
		return IContextMenuService.AFTER_GLOBAL_MENU_SECTION_ID;
	}

	/** Returns the class of the element that could be added to the library. */
	protected abstract <S extends INamedElement> Class<S> getElementClass();

	/**
	 * Returns the class of a library package that should contain elements of type given by
	 * {@link #getElementClass()}.
	 */
	protected abstract <S extends ILibraryPackage> Class<S> getLibraryPackageClass();

	/** Action to add an element to the library. */
	private class AddToLibraryAction extends Action {

		/** Element to be added to the library. */
		private EObject element;

		/** Constructor. */
		AddToLibraryAction(EObject element) {
			super("Add to Library", ToolingBaseUIActivator.getImageDescriptor("icons/library.png"));
			this.element = element;
		}

		/** {@inheritDoc} */
		@Override
		public void run() {

			String msg = checkLibrarySupported(element);
			if(msg != null) {
				showErrorAsync("Unsupported Operation", msg);
				return;
			}

			ILibraryPackage pack = openLibraryPackageDialog(getLibraryPackageClass());
			if(pack != null) {
				ModelElementLibraryService.getInstance().addElementToLibrary(pack, element);
			}
		}

		/**
		 * Checks if this element can be added to the library. Returns {@code null} if the element
		 * can be added to the library and an explanation {@link String} otherwise.
		 */
		protected String checkLibrarySupported(@SuppressWarnings("unused") EObject element) {
			return null;
		}
	}
}
