/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2013 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.compose;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.tooling.base.dnd.ConnectionDragContext;
import org.fortiss.tooling.base.dnd.ElementDropContext;
import org.fortiss.tooling.base.model.base.EntryConnectorBase;
import org.fortiss.tooling.base.model.base.ExitConnectorBase;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.kernel.extension.IConnectionCompositor;
import org.fortiss.tooling.kernel.extension.data.IConnectionCompositionContext;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.service.IElementCompositorService;

/**
 * Base class for all connection compositors.
 * 
 * @author igna
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 1A9F14BA9543BB14FE3501006F7C291B
 */
public abstract class ConnectionCompositorBase<HE extends IHierarchicElement, S extends EObject, T extends EObject>
		implements IConnectionCompositor<HE, S, T> {
	/** Creates the application specific {@link IConnection} model element. */
	protected abstract IConnection createConnection();

	/** Creates a connection and adds it to the parent element */
	protected IConnection createConnection(IConnector source, IConnector target, HE parent,
			ITopLevelElement topElement) {
		IConnection connection = createConnection();
		if(topElement != null) {
			topElement.prepareIDs(connection);
		}
		parent.getConnections().add(connection);

		boolean swap = false;

		if(source instanceof EntryConnectorBase) {
			if(target instanceof EntryConnectorBase) {
				swap = source.eContainer() != target.eContainer().eContainer();
			} else {
				swap = target instanceof ExitConnectorBase;
			}
		} else if(source instanceof ExitConnectorBase && target instanceof ExitConnectorBase) {
			swap = source.eContainer() == target.eContainer().eContainer();
		}
		connection.setSource(swap ? target : source);
		connection.setTarget(swap ? source : target);

		return connection;
	}

	/** {@inheritDoc} */
	@Override
	public boolean reconnect(EObject connection, S source, T target,
			IConnectionCompositionContext context) {
		// TODO (FH): https://af3.fortiss.org/issues/390
		// implement reconnect
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean canReconnect(EObject connection, S source, T target,
			IConnectionCompositionContext context) {
		// TODO (FH): https://af3.fortiss.org/issues/390
		// implement reconnect
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean canDisconnect(EObject connection) {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean disconnect(EObject connection) {
		if(connection instanceof IConnection) {
			// since an IConnection object contains no element, it can be safely deleted directly.
			EcoreUtil.delete(connection);
			return true;
		}
		return false;
	}

	/**
	 * Determines whether connectors of the component may be connected
	 * internally. The default value is <code>false</code>.
	 * 
	 * {@link #canConnectInternally} allows to state if we support internal
	 * connections. By default connectors of the parent element cannot be connected in internal view
	 * of the element.
	 * 
	 * @param source
	 * @param target
	 * @return whether connectors of the component may be connected internally.
	 */
	protected boolean canConnectInternally(S source, T target) {
		return false;
	}

	/** Creates a new source drop context */
	protected void createNewSourceDropContext(IHierarchicElement container,
			IConnectionCompositionContext context, EObject connected) {
		boolean isDragCtx = context instanceof ConnectionDragContext;
		ElementDropContext drop =
				isDragCtx ? ((ConnectionDragContext)context).getSourceDropContext() : null;
		IElementCompositorService.getInstance().compose(container, connected, drop);
	}

	/** Creates a new target drop context */
	protected void createNewTargetDropContext(IHierarchicElement container,
			IConnectionCompositionContext context, EObject contained) {
		boolean isDragCtx = context instanceof ConnectionDragContext;
		ElementDropContext drop =
				isDragCtx ? ((ConnectionDragContext)context).getTargetDropContext() : null;
		IElementCompositorService.getInstance().compose(container, contained, drop);
	}

	/**
	 * Returns true if the connector and the element to which the connection should be created have
	 * the same parent.
	 */
	protected boolean sameParent(IHierarchicElement hierarchicalElement, IConnector connector) {
		return hierarchicalElement.eContainer() == connector.eContainer();
	}

	/** Alters connector name if duplicate name already exists. */
	protected void avoidDuplicateConnectorName(IHierarchicElement container, IConnector connector) {
		String orig = ((INamedElement)connector).getName();
		int num;
		if(sameName(container, connector, orig)) {
			num = 1;
			while(sameName(container, connector, orig + num)) {
				num++;
			}
			((INamedElement)connector).setName(orig + num);
		} else {
			((INamedElement)connector).setName(orig);
		}
	}

	/** Checks whether the hierarchical element has connectors with a given name. */
	private boolean sameName(IHierarchicElement container, IConnector connector, String name) {
		for(IConnector sibling : container.getConnectors()) {
			if(sibling != connector && ((INamedElement)sibling).getName().equals(name)) {
				return true;
			}
		}
		return false;
	}
}
