/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.common.util;

import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Utility methods to ease the handling of Java 8 lambdas.
 * 
 * @author diewald
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 398AEE2F6FA7CB30D9B23BC49734AC46
 */
public class LambdaUtils {

	/** Constructor. */
	private LambdaUtils() {
		// Prevent instantiation.
	}

	/**
	 * Filters the given collection by the given filter and returns the remaining elements as a
	 * Collection.
	 * 
	 * @param collection
	 *            Collection to be filtered.
	 * @param filter
	 *            Filter to apply.
	 * @param sup
	 *            Supplier for the output {@link Collection}, e.g. TreeSet::new.
	 * @return Filtered collection as a Set.
	 */
	public static <T, S extends T> Collection<T> filter(Collection<S> collection,
			Predicate<S> filter, Supplier<? extends Collection<T>> sup) {
		return filterStream(collection, filter).collect(Collectors.toCollection(sup));
	}

	/**
	 * Filters the given collection by the given filter and returns the remaining elements as a
	 * Stream.
	 * 
	 * @param collection
	 *            Collection to be filtered.
	 * @param filter
	 *            Filter to apply.
	 * @return Filtered collection as a Stream.
	 */
	public static <S> Stream<S> filterStream(Collection<S> collection, Predicate<S> filter) {
		return collection.parallelStream().filter(filter);
	}

	/**
	 * Filters the given collection by the given filter and returns the remaining elements as a Set.
	 * 
	 * @param collection
	 *            Collection to be filtered.
	 * @param filter
	 *            Filter to apply.
	 * @return Filtered collection as a Set.
	 */
	public static <T, S extends T> Set<T> filterSet(Collection<S> collection, Predicate<S> filter) {
		return filterStream(collection, filter).collect(Collectors.toSet());
	}

	/**
	 * Filters the given collection by the given filter and returns the remaining elements as a
	 * List.
	 * 
	 * @param collection
	 *            Collection to be filtered.
	 * @param filter
	 *            Filter to apply.
	 * @return Filtered collection as a List.
	 */
	public static <T, S extends T> List<T>
			filterList(Collection<S> collection, Predicate<S> filter) {
		return filterStream(collection, filter).collect(Collectors.toList());
	}

	/**
	 * Filters the elements of the given {@code inCol} collection based on their type by the given
	 * {@code typeFilter}. The matching elements are added to the given {@code outCol} collection
	 * which is returned.
	 * 
	 * @param inCol
	 *            Collection to be filtered.
	 * @param outCol
	 *            Collection to which the filtered elements are added.
	 * @param typeFilter
	 *            Type for which the elements shall be filtered.
	 * @return Collection of elements of the given {@code filterType} (same as {@code outCol}).
	 */
	public static <S, T extends S, U extends T> Collection<T> filterByType(Collection<S> inCol,
			Collection<T> outCol, Class<U> typeFilter) {
		filterStream(inCol, e -> typeFilter.isAssignableFrom(e.getClass())).map(typeFilter::cast)
				.forEach(e -> outCol.add(e));
		return outCol;
	}

	/**
	 * Filters the given collection by the given filter and returns the first of the remaining
	 * elements .
	 * 
	 * @param collection
	 *            Collection to be filtered.
	 * @param filter
	 *            Filter to apply.
	 * @return First element of the filtered collection.
	 */
	@SuppressWarnings("unchecked")
	public static <T, S extends T> Optional<T> getFirst(Collection<S> collection,
			Predicate<S> filter) {
		return (Optional<T>)filterStream(collection, filter).findFirst();
	}

	/**
	 * Checks whether the given object inherits from any of the given Classes, i.e. the check
	 * {@link Class#isAssignableFrom(Class)} is performed for each given class on the given Object.
	 * 
	 * @param types
	 *            collection of classes to check inheritance.
	 * @param element
	 *            to check whether it inherits from any of the given classes.
	 * @return see above.
	 */
	public static boolean isAssignableFromAny(Collection<Class<?>> types, Object element) {
		return types.stream().anyMatch(t -> t.isAssignableFrom(element.getClass()));
	}

	/**
	 * Applies the mapping {@link Function} {@code mapper} to the given collection and returns the
	 * result as a collection whose implementation is given by the {@link Supplier}.
	 * 
	 * @param inColl
	 *            Collection of input elements.
	 * @param mapper
	 *            Mapping function.
	 * @param sup
	 *            Supplier for the output {@link Collection}, e.g. TreeSet::new.
	 * @return Collection of mapping results.
	 */
	public static <S, T> Collection<S> mapInOut(Collection<T> inColl, Function<T, S> mapper,
			Supplier<? extends Collection<S>> sup) {
		return inColl.parallelStream().map(mapper).collect(Collectors.toCollection(sup));
	}

	/**
	 * Applies the mapping {@link Function} {@code mapper} to the given collection and returns the
	 * result as a {@code Stream}.
	 * 
	 * @param inColl
	 *            Collection of input elements.
	 * @param mapper
	 *            Mapping function.
	 * @return Stream of the mapping result.
	 */
	public static <S, T> Stream<S> mapIn(Collection<T> inColl, Function<T, S> mapper) {
		return inColl.parallelStream().map(mapper);
	}

	/**
	 * Applies the mapping {@link Function} {@code mapper} to the given {@code stream} and returns
	 * the result as a collection whose implementation is given by the {@link Supplier}.
	 * 
	 * @param stream
	 *            Stream of input elements.
	 * @param mapper
	 *            Mapping function.
	 * @param sup
	 *            Supplier for the output {@link Collection}, e.g. TreeSet::new.
	 * @return Collection of mapping results.
	 */
	public static <S, T> Collection<S> mapOut(Stream<T> stream, Function<T, S> mapper,
			Supplier<? extends Collection<S>> sup) {
		return stream.map(mapper).collect(Collectors.toCollection(sup));
	}

	/**
	 * Applies the flat mapping {@link Function} {@code mapper} to the given collection and returns
	 * the result as a collection whose implementation is given by the {@link Supplier}.
	 * 
	 * @param inColl
	 *            Collection of input elements.
	 * @param mapper
	 *            Mapping function.
	 * @param sup
	 *            Supplier for the output {@link Collection}, e.g. TreeSet::new.
	 * @return Collection of mapping results.
	 */
	public static <S, T, R extends Stream<S>, U extends Collection<S>> Collection<S> flatMapInOut(
			Collection<T> inColl, Function<T, R> mapper, Supplier<U> sup) {
		return inColl.parallelStream().flatMap(mapper).collect(Collectors.toCollection(sup));
	}
}
