/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.ui.util;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;

/**
 * Utility methods for dealing with selections.
 * 
 * This class contains slight improvements with respect to
 * {@link org.conqat.ide.commons.ui.selection.SelectionUtils}.
 * 
 * @author ratiu
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: ACCFF3A4613831706A986FD737DC0E53
 */
public class SelectionUtils {

	/**
	 * For a selection provider object get the selected object, for simple
	 * selections a cast is performed to the desired type; for a structured
	 * selection returns the first element or <code>null</code> if the selection
	 * is empty. This returns <code>null</code> if the provided selection is
	 * null or the element is not of the required type.
	 * 
	 * @param selectionProvider
	 *            the selection provider
	 * @param type
	 *            the type of elements to be returned
	 * @return first element of the given type
	 */
	public static <T> T checkAndPickFirst(ISelectionProvider selectionProvider, Class<T> type) {
		return checkAndPickFirst(selectionProvider.getSelection(), type);
	}

	/**
	 * For simple selections a cast is performed to the desired type. For a
	 * structured selection this returns the first element or <code>null</code> if the selection is
	 * empty. This also returns <code>null</code> if the
	 * provided selection is <code>null</code> or the element is not of the
	 * required type.
	 * 
	 * @param selection
	 *            the current selection
	 * @param type
	 *            the type of elements to be returned
	 * @return the first element of the given type
	 */
	@SuppressWarnings("unchecked")
	public static <T> T checkAndPickFirst(ISelection selection, Class<T> type) {
		Object selectedElement;

		if(selection == null || selection.isEmpty())
			return null;

		if(!(selection instanceof IStructuredSelection)) {
			selectedElement = selection;
		} else {
			IStructuredSelection structuredSelection = (IStructuredSelection)selection;
			selectedElement = structuredSelection.getFirstElement();
		}

		if(!type.isInstance(selectedElement)) {
			return null;
		}
		return (T)selectedElement;
	}

	/**
	 * Returns a list of all the selected elements that have the given type or
	 * null if selection is null or is not an IStructuredSelection.
	 * 
	 * @param selection
	 *            the current selection
	 * @param type
	 *            the type of elements to be returned
	 * @return the list of elements of the given type
	 */
	@SuppressWarnings("unchecked")
	public static <T> List<T> selectAll(ISelection selection, Class<T> type) {
		if(selection == null || !(selection instanceof IStructuredSelection)) {
			return null;
		}

		IStructuredSelection structuredSelection = (IStructuredSelection)selection;
		List<T> elements = new ArrayList<T>();
		for(Object element : structuredSelection.toList()) {
			if(type.isInstance(element)) {
				elements.add((T)element);
			}
		}
		return elements;
	}

	/**
	 * Programmatically sets the objects to be selected in the currently active editor.
	 * 
	 * @param eObjects
	 *            {@link EObject}s to be selected.
	 */
	public static void setSelection(List<EObject> eObjects) {
		IWorkbenchWindow activeWorkbenchWindow =
				PlatformUI.getWorkbench().getActiveWorkbenchWindow();
		if(activeWorkbenchWindow == null) {
			return;
		}

		IWorkbenchPage activePage = activeWorkbenchWindow.getActivePage();
		if(activePage == null) {
			return;
		}

		IEditorPart activeEditor = activePage.getActiveEditor();
		if(activeEditor == null) {
			return;
		}

		IEditorSite editorSite = activeEditor.getEditorSite();
		if(editorSite == null) {
			return;
		}

		ISelectionProvider selectionProvider = editorSite.getSelectionProvider();
		if(selectionProvider == null) {
			return;
		}

		selectionProvider.setSelection(new StructuredSelection(eObjects));
	}
}
