/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.utils;

import static org.eclipse.emf.common.notify.Notification.EVENT_TYPE_COUNT;
import static org.eclipse.emf.ecore.util.EcoreUtil.getAllContents;
import static org.eclipse.emf.ecore.util.EcoreUtil.getRootContainer;
import static org.eclipse.emf.ecore.util.EcoreUtil.UsageCrossReferencer.find;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.impl.NotificationImpl;
import org.eclipse.emf.common.util.AbstractEList;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.ECollections;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EPackage.Registry;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.ecore.util.EcoreUtil.Copier;

/**
 * Utility methods for dealing with <code>Ecore</code> models. These methods should be used to
 * define EMF operations that make the models easier to use.
 * <P>
 * Implementations of these methods are tuned for efficiency in terms of not creating unnecessary
 * {@link EList}s.
 * 
 * @author ratiu
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 0A9636E007239C1CBBB229706A8CFB36
 */
public class EcoreUtils {

	/** The new event type count defined by the kernel. */
	public static final int KERNEL_EMF_EVENT_TYPE_COUNT = EVENT_TYPE_COUNT + 1;

	/**
	 * Notification to cause model element display refresh. We use a large number above EMF standard
	 * notifications to avoid collision with other EMF tools.
	 */
	private static final Notification refreshNotification = new NotificationImpl(
			EVENT_TYPE_COUNT + 1, null, null);

	/**
	 * Same as org.fortiss.tooling.kernel.utils.JavaUtils.convertList(Class<T>, List<S>) but with
	 * {@link EList} instead of {@link List}.
	 */
	@SuppressWarnings({"unchecked", "unused"})
	public static <S, T extends S> EList<T> convertList(Class<T> targetClass, EList<S> sourceList) {
		return (EList<T>)(EList<?>)sourceList;
	}

	/**
	 * <p>
	 * Converts an {@link EList} of a given type into an {@link EList} of one of its supertypes.
	 * </p>
	 * <p>
	 * Utility method to avoid unnecessary casts.
	 * </p>
	 * 
	 * @param targetClass
	 *            the class representing the super class.
	 * @param sourceList
	 *            the source list containing objects of type <code>S</code>, a subtype of
	 *            <code>T</code>.
	 * @return a target {@link EList} containing elements of type <code>targetClazz</code>.
	 */
	@SuppressWarnings("unchecked")
	public static <T, S extends T> EList<T>
			generalizeList(Class<T> targetClass, EList<S> sourceList) {
		return (EList<T>)(EList<?>)sourceList;
	}

	/**
	 * From a given {@link EList} with objects of type <code>S</code>, create another unmodifiable
	 * {@link EList} with objects of type <code>T</code>. The resulting {@link EList} is
	 * unmodifiable thereby it represents only a view over the source list.
	 * 
	 * @param target
	 *            a class representing type <code>T</code>.
	 * @param srcList
	 *            an EList with objects of type <code>S</code>.
	 * @return an unmodifiable EList of objects of type T.
	 */
	@SuppressWarnings("unchecked")
	public static <T> EList<T> pickInstanceOf(Class<T> target, List<?> srcList) {
		if(srcList == null) {
			return null;
		}
		EList<T> result = new BasicEList<T>();
		for(Object sourceElement : srcList) {
			if(target.isAssignableFrom(sourceElement.getClass())) {
				result.add((T)sourceElement);
			}
		}
		return ECollections.unmodifiableEList(result);
	}

	/**
	 * From a given {@link EList} with source objects of type <code>S</code> create another
	 * unmodifiable {@link EList} with objects of one of the types given by the parameter
	 * <code>targetClasses</code>, whereby those types are a sub-type of <code>S</code>. The
	 * resulting {@link EList} is unmodifiable thereby it represents only a view over the source
	 * list.
	 * 
	 * @param targetClasses
	 *            a collection of classes representing types of elements to be picked.
	 * @param sourceList
	 *            an EList with objects of type <code>S</code>.
	 * @return an unmodifiable EList of objects of type <code>S</code> containing only the
	 *         targetClasses instances.
	 */
	public static <S> EList<S>
			pickInstanceOf(Collection<Class<?>> targetClasses, List<S> sourceList) {
		if(sourceList == null) {
			return null;
		}
		EList<S> result = new BasicEList<S>();
		for(S sourceElement : sourceList) {
			Class<?> sourceClass = sourceElement.getClass();
			for(Class<?> targetClass : targetClasses) {
				if(targetClass.isAssignableFrom(sourceClass)) {
					result.add(sourceElement);
					break;
				}
			}
		}
		return ECollections.unmodifiableEList(result);
	}

	/**
	 * From a given {@link EList} with source objects of type <code>S</code> create another
	 * unmodifiable {@link EList} that DOES NOT contain objects of the type given by the parameter
	 * <code>targetClass</code>. The resulting {@link EList} is unmodifiable thereby it represents
	 * only a view over the source list.
	 * 
	 * @param targetClass
	 *            a class representing types of elements to be picked.
	 * @param sourceList
	 *            an {@link EList} with objects of type <code>S</code>.
	 * @return an unmodifiable {@link EList} of objects of type <code>S</code> containing only the
	 *         <code>targetClass</code> instances
	 */
	public static <S> EList<S> filterOutInstanceOf(Class<?> targetClass, List<S> sourceList) {
		if(sourceList == null) {
			return null;
		}
		EList<S> result = new BasicEList<S>();
		for(S sourceElement : sourceList) {
			Class<?> sourceClass = sourceElement.getClass();
			if(!targetClass.isAssignableFrom(sourceClass)) {
				result.add(sourceElement);
			}
		}
		return ECollections.unmodifiableEList(result);
	}

	/**
	 * From a given {@link EList} with source objects of type <code>S</code> pick the first object
	 * with type <code>T</code>, whereby <code>T</code> is a sub-type of <code>S</code>.
	 * 
	 * @param targetClass
	 *            a class representing type <code>T</code>.
	 * @param sourceList
	 *            an {@link EList} with objects of type <code>S</code>.
	 * @return an object of type <code>T</code> or <code>null</code> if none exists.
	 */
	@SuppressWarnings("unchecked")
	public static <S, T extends S> T pickFirstInstanceOf(Class<T> targetClass, List<S> sourceList) {
		if(sourceList == null) {
			return null;
		}
		for(S sourceElement : sourceList) {
			if(targetClass.isAssignableFrom(sourceElement.getClass())) {
				return (T)sourceElement;
			}
		}
		return null;
	}

	/**
	 * For a given {@link EObject} recursively returns all its children that have a
	 * certain type.
	 * 
	 * @param parent
	 *            the parent {@link EObject}.
	 * @param type
	 *            the type.
	 * @return a list of children.
	 */
	@SuppressWarnings("unchecked")
	public static <S> EList<S> getChildrenWithType(EObject parent, Class<S> type) {
		EList<S> children = new BasicEList<S>();
		TreeIterator<EObject> content = EcoreUtil.getAllContents(parent, true);

		while(content.hasNext()) {
			EObject child = content.next();
			if(type.isAssignableFrom(child.getClass())) {
				children.add((S)child);
			}
		}
		return children;
	}

	/**
	 * For a given {@link EObject}, recursively returns the first child that has a certain type.
	 * Using this method is more efficient than returning the first element of the result of
	 * {@link #getChildrenWithType(EObject, Class)}.
	 * 
	 * @param parent
	 *            the parent {@link EObject}
	 * @param type
	 *            the type
	 * 
	 * @return the first child of the given type.
	 */
	@SuppressWarnings("unchecked")
	public static <S> S getFirstChildWithType(EObject parent, Class<S> type) {
		TreeIterator<EObject> content = EcoreUtil.getAllContents(parent, true);

		while(content.hasNext()) {
			EObject child = content.next();
			if(type.isAssignableFrom(child.getClass())) {
				return (S)child;
			}
		}
		return null;
	}

	/**
	 * Returns a list of all parent {@link EObject}s that have the type <code>type</code>. If none
	 * are found, an empty list is returned.
	 * 
	 * @param startElement
	 *            The {@link EObject} from which the "upwards" search shall begin.
	 * @param type
	 *            The class type which is used to filter the search.
	 * @return List of parent {@link EObject}s of type <code>type</code>.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends EObject> EList<T> getParentsWithType(EObject startElement,
			Class<T> type) {

		EList<T> parentElements = new BasicEList<T>();
		EObject currentParent = startElement.eContainer();

		while(currentParent != null) {
			if(type.isAssignableFrom(currentParent.getClass())) {
				parentElements.add((T)currentParent);
			}
			currentParent = currentParent.eContainer();
		}
		return parentElements;
	}

	/**
	 * Returns the first parent {@link EObject} that has the type <code>T</code>. If none are
	 * found, null is returned.
	 * 
	 * @param startElement
	 *            The {@link EObject} from which the "upwards" search shall begin.
	 * @param type
	 *            The class type to find the first parent with.
	 * @return List of parent {EObject} of type <code>T</code>.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends EObject> T getFirstParentWithType(EObject startElement, Class<T> type) {
		EObject currentParent = startElement.eContainer();
		while(currentParent != null) {
			if(type.isAssignableFrom(currentParent.getClass())) {
				return (T)currentParent;
			}
			currentParent = currentParent.eContainer();
		}
		return null;
	}

	/**
	 * Returns true if <code>candidateAncestor</code> is an ancestor of <code>startElement</code>.
	 * 
	 * An element is not an ancestor of itself!
	 */
	public static boolean isAncestor(EObject startElement, EObject candidateAncestor) {
		EObject currentParent = startElement.eContainer();
		while(currentParent != null) {
			if(currentParent == candidateAncestor) {
				return true;
			}
			currentParent = currentParent.eContainer();
		}
		return false;
	}

	/**
	 * Posts a {@link Notification} with an event type ID outside the EMF
	 * default notification IDs to cause UI refreshes.
	 * 
	 * @param eObject
	 *            the target model element to be refreshed
	 */
	public static void postRefreshNotification(EObject eObject) {
		eObject.eNotify(refreshNotification);
	}

	/**
	 * Hack to the EcoreUtils copier which removes some duplicates in some lists, whereas we DO NOT
	 * want it.
	 */
	private static class HackedCopier extends EcoreUtil.Copier {
		/** {@inheritDoc} */
		@Override
		protected void copyContainment(EReference eReference, EObject eObject, EObject copyEObject) {
			if(eObject.eIsSet(eReference)) {
				if(eReference.isMany()) {
					@SuppressWarnings("unchecked") List<EObject> source =
							(List<EObject>)eObject.eGet(eReference);
					@SuppressWarnings("unchecked") List<EObject> target =
							(List<EObject>)copyEObject.eGet(getTarget(eReference));
					if(source.isEmpty()) {
						target.clear();
					} else {
						addAll(target, copyAll(source));
					}
				} else {
					EObject childEObject = (EObject)eObject.eGet(eReference);
					copyEObject.eSet(getTarget(eReference), childEObject == null ? null
							: copy(childEObject));
				}
			}
		}
	}

	/**
	 * Hacked <code>addAll</code< method to circumvent the removal of duplicate elements.
	 * 
	 * In theory this should not be necessary: if a list requires non duplicate elements, then the
	 * default of removing these duplicates when adding to a list should be kept. The problem is
	 * that (since Kepler?) EMF forces us to set some lists as "unique" when we do not want it to
	 * (typically the arguments of a function in an expression). So since we do not have any other
	 * choice at the level of the meta-model, we have to hack. Here is the hack.
	 */
	public static <T extends EObject, U extends T> void addAll(List<T> l1, Collection<U> l2) {
		if(l1 instanceof AbstractEList<?>) {
			// Here is the hack: we use addAllUnique instead of addAll
			((AbstractEList<T>)l1).addAllUnique(l2);
		} else {
			// In other cases we just revert to the standard behavior
			l1.addAll(l2);
		}
	}

	/**
	 * Returns a self-contained copy of the given <code>eObject</code>.
	 * 
	 * @param eObject
	 *            the object to copy.
	 * @return the copy.
	 * @see Copier
	 */
	public static <T extends EObject> T copy(T eObject) {
		Copier copier = new HackedCopier();
		EObject result = copier.copy(eObject);
		copier.copyReferences();

		@SuppressWarnings("unchecked") T t = (T)result;
		return t;
	}

	/** Map to cache potentially expensive lookups in {@link #getEClassForClass(Class)}. */
	private static Map<Class<?>, EClass> clazz2EClassCache = new HashMap<Class<?>, EClass>();

	/**
	 * <p>
	 * Determines the {@link EClass} corresponding to a give {@code clazz} (typically the interface
	 * for the generated implementation class of an Ecore meta-model), or returns {@code null} if no
	 * such {@link EClass} could be found.
	 * </p>
	 * <p>
	 * <b>CAVEAT:</b> For each {@code clazz}, the first invocation of this method is potentially
	 * expensive.
	 * </p>
	 */
	public static EClass getEClassForClass(Class<?> clazz) {
		EClass eClass = null;

		if(clazz2EClassCache.containsKey(clazz)) {
			return clazz2EClassCache.get(clazz);
		}

		// Query all registered EPackages for the given clazz
		Registry registry = EPackage.Registry.INSTANCE;
		for(String key : new HashSet<String>(registry.keySet())) {
			EPackage ePackage;
			try {
				ePackage = registry.getEPackage(key);
			} catch(Exception e) {
				System.out
						.println("Error: EPackage for \"" + key +
								"\" has not been registered (in extension point \"org.eclipse.emf.ecore.generated_package\").");
				continue;
			}
			for(EClassifier eClassifier : ePackage.getEClassifiers()) {
				if(eClassifier instanceof EClass && eClassifier.getInstanceClass() != null &&
						clazz.isAssignableFrom(eClassifier.getInstanceClass())) {
					eClass = (EClass)eClassifier;
					clazz2EClassCache.put(clazz, eClass);
					return eClass;
				}
			}
		}

		return null;
	}

	/** Computes the level of an {@link EObject} in the tree. */
	public static int getModelElementLevel(EObject o) {
		int level = 0;
		while(o.eContainer() != null) {
			level++;
			o = o.eContainer();
		}

		return level;
	}

	/**
	 * Returns the ancestor of the {@link EObject} {@code o} that is {@code levelsUp} levels above
	 * model element in the hierarchy, or {@code null}, in case {@code levelsUp < 0} or
	 * {@code levelsUp > depth(modelElement)}.
	 */
	public static EObject getModelElementAncestor(EObject o, int levelsUp) {
		if(levelsUp < 0) {
			return null;
		}
		while(o.eContainer() != null && levelsUp > 0) {
			o = o.eContainer();
			levelsUp--;
		}
		return levelsUp == 0 ? o : null;
	}

	/**
	 * Retrieves all EStructuralFeature.Setting, which are a reference to the given
	 * 'referencedEObject'.
	 * 
	 * Note: This implementation is adopted from {@link EcoreUtil}.delete().
	 * 
	 * @param referencedEObject
	 *            The {@link EObject}, to which all references are supposed to be retrieved.
	 * @return A {@link Collection}, containg all EStructuralFeature.Setting>, which are references
	 *         to 'referencedEObject'.
	 */
	public static Collection<EStructuralFeature.Setting>
			getAllReferences(EObject referencedEObject) {
		EObject rootEObject = getRootContainer(referencedEObject);
		Resource resource = rootEObject.eResource();

		Collection<EStructuralFeature.Setting> usages;
		if(resource == null) {
			usages = find(referencedEObject, rootEObject);
		} else {
			ResourceSet resourceSet = resource.getResourceSet();
			usages =
					(resourceSet == null) ? find(referencedEObject, resource) : find(
							referencedEObject, resourceSet);
		}
		return usages;
	}

	/**
	 * <p>
	 * Disables the notifications emitted by the given {@link EObject} and its children.
	 * </p>
	 * <p>
	 * <b>CAVEAT:</b> This method should be only used to avoid a notification flood when editing
	 * models massively. The Undo/Redo-logging is disabled thereby. You must ensure to re-enable the
	 * notifications, after the modifications are done.
	 * </p>
	 * 
	 * @param element
	 *            The element for which and its children the notifications shall be disabled.
	 * 
	 * @return {@code true}, if the notifications can be disabled for the given {@link EObject}
	 */
	public static boolean disableNotificationsOf(EObject element) {
		return setNotificationsOf(element, false);
	}

	/**
	 * Enables the notifications for the given {@link EObject} and its children.
	 * 
	 * @param element
	 *            The element for which and its children the notifications shall be enabled.
	 * 
	 * @return {@code true}, if the notifications can be disabled for the given {@link EObject}
	 */
	public static boolean enableNotificationsOf(EObject element) {
		return setNotificationsOf(element, true);
	}

	/**
	 * Sets the notification state for the given {@link EObject} and its children.
	 * 
	 * @param element
	 *            The element for which and its children the notification state should be.
	 * 
	 * @param value
	 *            Whether to enable or disable notifications of the given {@code element}.
	 * 
	 * @return {@code true}, if the notifications can be disabled for the given {@link EObject}
	 */
	private static boolean setNotificationsOf(EObject element, boolean value) {
		element.eSetDeliver(value);
		Iterator<EObject> childIterator = getAllContents(element, true);
		while(childIterator.hasNext()) {
			EObject currentChild = childIterator.next();
			currentChild.eSetDeliver(value);
		}
		return true;
	}

	/**
	 * @param s
	 * @return an {@link EList} containing the elements of <code>s</code>.
	 */
	public static <T> EList<T> collectToEList(Stream<T> s) {
		return new BasicEList<T>(s.collect(Collectors.toList()));
	}
}
